/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvimage.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIMAGE_H
#define QVIMAGE_H

#include <QMetaType>
#include <qvcore/qvimagebuffer.h>

#define	QVIMAGE_INIT_READ(TYPE, IMAGE)					\
	const TYPE * __qv_data_##IMAGE##__ = IMAGE.getReadData();	\
	const uInt __qv_step_##IMAGE##__ = IMAGE.getStep();		\
	const uChar	__qv_size_##IMAGE##__ = sizeof(TYPE),		\
			__qv_planes_##IMAGE##__  = IMAGE.getPlanes();

#define	QVIMAGE_INIT_WRITE(TYPE, IMAGE)					\
	TYPE * __qv_data_##IMAGE##__ = IMAGE.getWriteData();		\
	const uInt __qv_step_##IMAGE##__ = IMAGE.getStep();		\
	const uChar	__qv_size_##IMAGE##__ = sizeof(TYPE),		\
			__qv_planes_##IMAGE##__  = IMAGE.getPlanes();

#define	QVIMAGE_PTR_INIT_READ(TYPE, IMAGE)				\
	const TYPE * __qv_data_##IMAGE##__ = IMAGE->getReadData();	\
	const uInt __qv_step_##IMAGE##__ = IMAGE->getStep();		\
	const uChar	__qv_size_##IMAGE##__ = sizeof(TYPE),		\
			__qv_planes_##IMAGE##__  = IMAGE->getPlanes();

#define	QVIMAGE_PTR_INIT_WRITE(TYPE, IMAGE)				\
	TYPE * __qv_data_##IMAGE##__ = IMAGE->getWriteData();		\
	const uInt __qv_step_##IMAGE##__ = IMAGE->getStep();		\
	const uChar	__qv_size_##IMAGE##__ = sizeof(TYPE),		\
			__qv_planes_##IMAGE##__  = IMAGE->getPlanes();

#define QVIMAGE_PIXEL(IMAGE, Col, Row, Channel)	\
	(__qv_data_##IMAGE##__ [(Row)* __qv_step_##IMAGE##__ /__qv_size_##IMAGE##__ + __qv_planes_##IMAGE##__ *(Col)+(Channel)])

/// @class QVGenericImage qvimage.h qvcore/qvimage.h
/// @brief QVGenericImage class.
///
/// This class defines generic functionality for images, and shares it with every image, due to it's parental relation
/// (from a inheritance point of view) with class QVImage.
/// QVImage will specialize from this class, and add specific functions to access image content mainly, for a concrete bit depth
/// and number of channels.
///
/// @ingroup qvcore
class QVGenericImage
	{
	public:
		/// @brief Default constructor.
		///
		/// Initializes image ROI and anchor to zero offset and width, both.
		QVGenericImage (): roi(0,0,0,0), anchor(QPoint(0,0))					{ }

		/// @brief Copy constructor.
		///
		/// Initializes image ROI and anchor to ROI and anchor from image 'img'.
		QVGenericImage(QVGenericImage const &img): roi(img.getROI()), anchor(img.getAnchor())	{ }

		/// @brief Virtual destructor for QVGenericImage.
		virtual ~QVGenericImage ()								{ };

		/// @brief Virtual method for getting number of columns from image.
		///
		/// @returns number of columns for image,
		virtual uInt	getCols()		const	= 0;

		/// @brief Virtual method for getting number of rows from image.
		///
		/// @returns number of rows for image,
		virtual uInt	getRows()		const	= 0;

		/// @brief Virtual method for getting step from image.
		///
		/// This functions returns the distance in bytes between starts of consecutive lines in the image.
		///
		/// This may be not equal to a multiple of the number of columns, so this value is necessary to normal access
		/// to the image.
		///
		/// @returns distance in bytes between two consecutive lines in the image.
		virtual uInt	getStep()		const	= 0;

		/// @brief Virtual method for getting number of planes for the image.
		///
		/// Returned value should be equal to template parameter 'Planes', for any QVImage. This is the number
		/// of channels that can be accessed in the image.
		///
		/// @returns number of channels of the image.
		virtual uInt	getPlanes()		const	= 0;

		/// @brief Virtual method for getting bit depth of the image, in bytes.
		///
		/// Returned value should be equal to size of template parameter 'Type', for any QVImage. This is the size
		/// of any element of the data array that stores the image in memory.
		///
		/// @returns number of channels of the image.
		virtual	uInt	getTypeSize()		const	= 0;

		/// @brief Virtual method for getting data size for the image.
		///
		/// This functions, for any image, returns the following value:
		/// \verbatim dataSize = rows * step * planes \endverbatim
		///
		/// @returns size in bytes of data buffer for the image.
		virtual	uInt	getDataSize()		const	= 0;

		/// @brief Method for obtaining actual region of interest for image.
		///
		/// The region of interest of an image is used for various operations over it. Check the manual for further
		/// documentation
		///
		/// @returns a QRect that contains the region of interest of the image.
		const QRect &	getROI()		const	{ return roi; }

		/// @brief Obtains actual anchor for image.
		///
		/// The anchor of an image is used for various operations over it. Check the manual for further
		/// documentation
		///
		/// @returns a QPoint that indicates anchor of the image.
		const QPoint &	getAnchor()		const	{ return anchor; }

		// ROI operations
		/// @brief Resets the region of interest of the image.
		///
		/// Makes the region of interest of the image hold the whole image.
		///
		/// That is, makes the region of interest be the rectangle which has upper left corner in pixel (0,0),
		/// and lower right corner in pixel (columns, rows), where columns and rows are the number of columns
		/// and rows of the image.
		///
		void resetROI()				{ setROI(0,0,getCols(), getRows()); }

		/// @brief Reduces the region of interest of the image, in a given amount of pixels.
		///
		/// Adds cols and rows parameters to the top left corner of the region of interest, and those values
		/// doubled, to the width.
		///
		/// That is, reduces the region of interest in a border, of columns and rows indicated in the parameters.
		///
		/// @param cols	columns which should be reduced from the border of the roi.
		/// @param rows	rows which should be reduced from the border of the roi.
		void erodeROI(uInt cols, uInt rows)
			{ setROI(getROI().x()+cols, getROI().y()+rows, getROI().width()-2*cols, getROI().height()-2*rows); }

		/// @brief Increments the region of interest of the image, in a given amount of pixels.
		///
		/// Substracts cols and rows parameters to the top left corner of the region of interest, and those values
		/// doubled, to the width.
		///
		/// That is, increments the region of interest in a border, of columns and rows indicated in the parameters.
		///
		/// @param cols	columns which should be incremented to the border of the roi.
		/// @param rows	rows which should be incremented to the border of the roi.
		void dilateROI(uInt cols, uInt rows)
			{ setROI(getROI().x()-cols, getROI().y()-rows, getROI().width()+2*cols, getROI().height()+2*rows); }

		/// @brief Sets dimensions and location of region of interest in an image.
		///
		/// Region of interest for an image is a rectangle significant to many operations that can be done over an
		/// image.
		///
		/// It is defined by pixel where top left corner for this rectangle is located, and width & height of
		/// this rectangle.
		///
		/// @param x column for the pixel where top left corner of the region of interest is located.
		/// @param y row for the pixel where top left corner of the region of interest is located.
		/// @param w width for the region of interest.
		/// @param h height for the region of interest.
		void setROI(int x, int y, uInt w, uInt h) { setROI(QRect(x,y,w,h)); }

		/// @brief Sets the region of interest for an image, holding the whole image, except a border of a given width.
		///
		/// @param margin size of the border.
		void setMarginROI(int margin)
			{ setROI(margin, margin, getCols() - 2*margin, getRows() -2*margin); }

		/// @brief Directly sets the region of interest for an image, with a given rectangle.
		///
		/// @param rect rectangle that represents region of interest.
		void setROI(const QRect &rect)
			{
			Q_ASSERT_X(rect.x() >= 0,"QVGenericImage::setROI()","QRect.x() is less than zero");
			Q_ASSERT_X(rect.y() >= 0,"QVGenericImage::setROI()","QRect.y() is less than zero");
			Q_ASSERT_X(rect.width() > 0,"QVGenericImage::setROI()","QRect.width() is less or equal to zero");
			Q_ASSERT_X(rect.height() > 0,"QVGenericImage::setROI()","QRect.height() is less or equal to zero");
			Q_ASSERT_X(rect.x()+rect.width() <= (int) getCols(),"QVGenericImage::setROI()","x + width > columns");
			Q_ASSERT_X(rect.x()+rect.height() <= (int) getRows(),"QVGenericImage::setROI()","y + height > rows");
			roi = rect;
			}

		// Anchor operations
		/// @brief Sets anchor to pixel (0,0).
		///
		void resetAnchor()			{ setAnchor(0,0); }

		/// @brief Sets anchor to a given pixel coordinates.
		///
		/// @param col column of the pixel where the anchor is located.
		/// @param row row of the pixel where the anchor is located.
		void setAnchor(int col, int row)	{ setAnchor(QPoint(col,row)); }

		/// @brief Directly sets anchor to a given pixel coordinates, given a point.
		///
		/// @param point QPoint where the anchor is located.
		void setAnchor(const QPoint &point)
			{
			Q_ASSERT_X(point.x()>=0,"QVGenericImage::setAnchor()","horizontal value for anchor is less than zero");
			Q_ASSERT_X(point.y()>=0,"QVGenericImage::setAnchor()","vertical value for anchor is less than zero");
			Q_ASSERT_X(point.x() < (int) getCols(),"QVGenericImage::setAnchor()","horizontal value exceeds cols");
			Q_ASSERT_X(point.y() < (int) getRows(),"QVGenericImage::setAnchor()","vertical value exceeds rows");
			anchor = point;
			}

		/// @brief Function to make introspection in the image, and obtain a char string identifying the type of the
		/// real image contained in the QVGenericImage.
		///
		/// @returns pointer to a constant char string identifying image type.
		virtual const char * getTypeQString() const = 0;

		/// @brief Function to check whether an image is compatible with given image type id char string.
		///
		/// This functions checks if this image has same bit depth and channel number, as the type of QVGenericImage
		/// represented by the parameter string.
		///
		/// Id char string can be that returned by getTypeQString() function, from another QVGenericImage, or
		/// one specified by the programmer.
		///
		/// This function is used to dynamic type introspection, to determine subclass of a QVGenericImage, and
		/// perform different functionality with it, depending on bit depth or number of planes. An example code for
		/// this use is below:
		/// \code [...]
		/// QVImage image;
		///
		/// [...]
		///
		/// if(image->isCompatibleWith("QVImage<uChar,1>"))
		///	// do stuff with the uChar bit depth, one channel image ...
		/// else if(image->isCompatibleWith("QVImage<uChar,3>"))
		///	// do stuff with the uChar bit depth, three channel image ...
		/// [...]
		/// else
		///	// alternate code for that type of image ...
		///
		/// [...] \endcode
		///
		/// @param qvimageClassName char string identifying QVGenericImage type of an image.
		/// @returns true if this image is compatible in type with that represented by the string.
		bool isCompatibleWith(const char *qvImageClassName) const
			{
			Q_ASSERT_X(qvImageClassName!= NULL,"QVGenericImage::isCompatibleWith()","class name string is NULL");
			return (0 == strcmp(this->getTypeQString(),qvImageClassName));
			}

		/// @brief Function to check whether an image is compatible with this image.
		///
		/// This functions checks if this image is compatible with a given image (that is, it has same bit depth and
		/// same number of channels).
		///
		/// Id char string can be that returned by getTypeQString() function, from another QVGenericImage, or
		/// one specified by the programmer.
		///
		/// @param image image to check type with.
		/// @returns true if this image is compatible in type with that represented by the string.
		bool isCompatibleWith(const QVGenericImage *image) const
			{
			Q_ASSERT_X(image!= NULL,"QVGenericImage::isCompatibleWith()","NULL pointer");
			return this->isCompatibleWith(image->getTypeQString());
			}

	protected:
		QRect roi;
		QPoint anchor;
	};

/// @class QVGenericImage qvimage.h qvcore/qvimage.h
/// @brief QVImage class.
///
/// This class extends QVGenericImage, and concretes bit depth and channel number content access functionality.
///
/// @ingroup qvcore
template <typename Type, int Planes = 1> class QVImage: public QVGenericImage 
	{
	public:
		/// @brief Default constructor.
		///
		/// Initializes image with a data buffer of size 1x1. Sets anchor to pixel (0,0), and region of interest
		/// to a rectangle containing the only pixel of the image.
		///
		QVImage():QVGenericImage()
			{
			this->imageBuffer = new QVImageBuffer<Type, Planes>(1, 1);
			setROI(0,0, this->imageBuffer->getCols(), this->imageBuffer->getRows());
			setAnchor(0,0);
			}

		/// @brief Dimensions and buffer constructor.
		///
		/// Creates an image of dimension cols x rows, and optionally you can define a given step for the image.
		///
		/// If buffer points to a region of memory, it will use that region as image data.
		///
		/// DEPRECATED: you shouldn't be able to specify a pointer to a buffer.
		/// If you need a image data buffer, create an image, and use it's.
		///
		/// @param cols number of columns for the image.
		/// @param rows number of rows for the image.
		/// @param step defined step for the image.
		/// @param buffer pointer to a memory zone where image is and will be stored.
		QVImage(uInt cols, uInt rows, uInt step = 0, const Type * buffer = NULL):QVGenericImage()
			{
			this->imageBuffer = new QVImageBuffer<Type, Planes>(cols, rows, step, buffer);
			setROI(0,0, cols, rows);
			setAnchor(0,0);
			}

		/// @brief Copy constructor
		///
		/// This constructor can be used with images of same bit depth, or which have same channel number, than those
		/// given by parameter.
		///
		/// In this case, this constructor will be defined for images of uChar bit depth, or that have just one
		/// channel number.
		///
		/// If given QVImage has same bit depth and plane numbers, this constructor will copy the reference to
		/// that image data buffer, and will use copy on write technique if needed.
		/// Otherwise, it will allocate a new data buffer for the image, and make a conversion from different bit
		/// depth or channel number.
		///
		/// The region of interest, and the anchor of this image will be set to those from the image given by parameter.
		///
		/// @param img image to be copied.
		QVImage(QVImage<uChar,1> const &img);

		/// @brief Overloaded copy constructor.
		QVImage(QVImage<uChar,3> const &img);

		/// @brief Overloaded copy constructor.
		QVImage(QVImage<sShort,1> const &img);

		/// @brief Overloaded copy constructor.
		QVImage(QVImage<sShort,3> const &img);

		/// @brief Overloaded copy constructor.
		QVImage(QVImage<sFloat,1> const &img);

		/// @brief Overloaded copy constructor.
		QVImage(QVImage<sFloat,3> const &img);
		
		/// @brief Overloaded function from QVGenericImage::getTypeQString()
		const char * getTypeQString() const;

		/// @brief Overloaded function from QVGenericImage::getRows()
		uInt getRows()		const	{ return imageBuffer->getRows(); }

		/// @brief Overloaded function from QVGenericImage::getCols()
		uInt getCols()		const	{ return imageBuffer->getCols(); }

		/// @brief Overloaded function from QVGenericImage::getStep()
		inline uInt getStep()	const	{ return imageBuffer->getStep(); }

		/// @brief Overloaded function from QVGenericImage::getPlanes()
		inline uInt getPlanes()	const	{ return imageBuffer->getPlanes(); }

		/// @brief Overloaded function from QVGenericImage::getDataSize()
		uInt getDataSize()	const	{ return imageBuffer->getDataSize(); }

		/// @brief Overloaded function from QVGenericImage::getTypeSize()
		uInt getTypeSize()	const	{ return imageBuffer->getTypeSize(); }

		/// @brief Method to obtain image data buffer, in read mode
		///
		/// This method can be used to program fast access algorithms to pixel values in an image.
		///
		/// It is intended to be used only to read pixels in an image, thus getWriteData() method should be used
		/// when access to write or read/write is intended to be done.
		///
		/// This function will generally be faster than getWriteData(), because it will not need to realize copy on write,
		/// in the case that the image data buffer where shared with other QVImage.
		///
		const Type * getReadData()	const	{ return imageBuffer->getReadData(); }

		/// @brief Method to obtain image data buffer, in read/write mode
		///
		/// This method can be used to program fast access algorithms to pixel values in an image.
		///
		/// This function will perform copy on write in the case that the data buffer for this image where shared with
		/// other QVImage's.
		///
		/// Thus can be slower than getReadData() method to access image pixels, but will ensure avoiding side
		/// effects on modifying shared buffers with other images.
		///
		Type * getWriteData()		{ return imageBuffer->getWriteData(); }

		/// @brief Sets pixel values for an image, to a given value.
		///
		/// This method  uses the region of interest of the image, to set pixels inside it to a given value.
		/// It can be used either for one channel, or several channel images.
		///
		/// In the former case, it should be used with one parameter, only specifying value for the first channel,
		/// otherwise, values for other channels will be ignored.
		/// If no parameters are given, the image region of interest will be set to zero value pixels.
		///
		/// In the latter case, default values of 0 will be used if there are unspecified channel values.
		///
		/// @param c1 value to set pixel's first channel in the region of interest of the image.
		/// @param c2 value to set pixel's second channel in the region of interest of the image.
		/// @param c3 value to set pixel's third channel in the region of interest of the image.
		void set(Type c1 = 0, Type c2 = 0, Type c3 = 0);

		/// @brief Operator to access pixel values for a pixel channel, given it's coordinates.
		///
		/// This operator can be used either with one channel or multichannel images.
		/// In the first case, it shouldn't be given a channel number, to use the default zero value.
		///
		/// @param col column number of pixel to be accessed.
		/// @param row row number of pixel to be accessed.
		/// @param channel channel number of pixel to be accessed.
		inline Type &operator()(const uInt col, const uInt row, const uInt channel = 0)
			{
			Type * data = imageBuffer->getWriteData();
			uInt step = getStep();
			//int dataSize = imageBuffer->getDataSize();

			Q_ASSERT_X(col < getCols(),"QVImage::operator()","col further upper bound");
			Q_ASSERT_X(row < getRows(),"QVImage::operator()","row further upper bound");
			int idx = row*step/sizeof(Type) + Planes*col+channel;
			Q_ASSERT_X(idx >= 0,"QVImage::operator()","accessing below data");
			Q_ASSERT_X(idx < imageBuffer->getDataSize(),"QVImage::operator()","accessing above data");
			return data[idx];
			}

		inline Type operator()(const uInt col, const uInt row, const uInt channel = 0) const
			{
			Type const * data = imageBuffer->getReadData();
			uInt cols = getCols(), rows = getRows();
			uInt step = getStep();
			int dataSize = getDataSize();

			Q_ASSERT_X(col < cols,"QVImage::operator()","row out of upper bound");
			Q_ASSERT_X(row < rows,"QVImage::operator()","row out of upper bound");
			int idx = row*step/sizeof(Type) + Planes*col+channel;
			Q_ASSERT_X(idx >= 0,"QVImage::operator()","accessing below data");
			Q_ASSERT_X(idx < dataSize,"QVImage::operator()","accessing above data");
			return data[idx];
			}

		/// @brief Overload of operator (), to access pixel values for a pixel channel, given a QVPoint and a
		/// channel number.
		///
		/// This overload is like the previous, but can be used with a QPoint, to access a concrete pixel in the image.
		/// @param point QPoint indicating location of the pixel to be accessed.
		/// @param channel channel number of pixel to be accessed.
		inline Type &operator()(const QPoint point, const uInt channel = 0)
			{
			Type * data = imageBuffer->getWriteData();
			uInt cols = getCols(), rows = getRows(), step = getStep();
			int dataSize = getDataSize();

			Q_ASSERT_X(point.x() < (int) cols,"QVImage::operator()","col further upper bound");
			Q_ASSERT_X(point.y() < (int) rows,"QVImage::operator()","row further upper bound");
			int idx = point.y()*step/sizeof(Type) + Planes*point.x()+channel;
			Q_ASSERT_X(idx >= 0,"QVImage::operator()","accessing below data");
			Q_ASSERT_X(idx < dataSize,"QVImage::operator()","accessing above data");
			return data[idx];
			}

		inline Type operator()(const QPoint point, const uInt channel = 0) const
			{
			Type const * data = imageBuffer->getReadData();
			uInt cols = getCols(), rows = getRows(), step = getStep();
			int dataSize = getDataSize();

			Q_ASSERT_X(point.x() < (int) cols,"QVImage::operator()","row out of upper bound");
			Q_ASSERT_X(point.y() < (int) rows,"QVImage::operator()","row out of upper bound");
			int idx = point.y()*step/sizeof(Type) + Planes*point.x()+channel;
			Q_ASSERT_X(idx >= 0,"QVImage::operator()","accessing below data");
			Q_ASSERT_X(idx < dataSize,"QVImage::operator()","accessing above data");
			return data[idx];
			}

		/// @brief Copy operator
		///
		/// This operator can be used with images of same bit depth, or which have same channel number, than those
		/// given by parameter.
		///
		/// In this case, this operator will be defined for images of uChar bit depth, or that have just one
		/// channel number.
		///
		/// If given QVImage has same bit depth and plane numbers, this operator will copy the reference to
		/// that image data buffer, and will use copy on write technique if needed.
		/// Otherwise, it will allocate a new data buffer for the image, and make a conversion from different bit
		/// depth or channel number.
		///
		/// The region of interest, and the anchor of this image will be set to those from the image given by parameter.
		///
		/// @param img image to be copied.
		QVImage<Type, Planes> & operator=(const QVImage<uChar, 1> &sourceImage);

		/// @brief Overloaded copy operator.
		QVImage<Type, Planes> & operator=(const QVImage<uChar, 3> &sourceImage);

		/// @brief Overloaded copy operator.
		QVImage<Type, Planes> & operator=(const QVImage<sShort, 1> &sourceImage);

		/// @brief Overloaded copy operator.
		QVImage<Type, Planes> & operator=(const QVImage<sShort, 3> &sourceImage);

		/// @brief Overloaded copy operator.
		QVImage<Type, Planes> & operator=(const QVImage<sFloat, 1> &sourceImage);

		/// @brief Overloaded copy operator.
		QVImage<Type, Planes> & operator=(const QVImage<sFloat, 3> &sourceImage);
		
		/// @brief Equality compare operator
		///
		/// This operator will compare two images of same bit depth and channel number, and will return true if
		/// images are equal in all of these:
		///	- dimensions (cols and rows).
		///	- region of interest.
		///	- content of region of interest.
		/// returning false otherwise.
		///
		/// @param img image to compare with this image.
		bool operator==(const QVImage<Type, Planes> &img) const;

		/// @brief Inequality compare operator
		///
		/// This operator will compare two images of same bit depth and channel number, and will return true if
		/// images differ in any of:
		///	- dimensions (cols and rows).
		///	- region of interest.
		///	- content of region of interest.
		/// returning false otherwise.
		///
		/// DEPRECATED: Shouldn't ignore dimensions in caparisons, and only check if region of interest content are
		/// the same?
		///
		/// @param img image to compare with this image.
		/// @returns true if images are practically equal (same content of region of interest and dimensions).
		bool operator!=(const QVImage<Type, Planes> &img) const { return !(*this == img); }

		/// @brief Pixel wise 'less than' compare operator
		///
		/// @param img image to compare with this image.
		/// @returns 1-channel uChar bit depth image containing non-zero values for pixels which value for corresponding
		/// pixel in this image is less than value for corresponding pixel in given image.
		QVImage<uChar> operator<(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise 'greater than' compare operator
		///
		/// @param img image to compare with this image.
		/// @returns 1-channel uChar bit depth image containing non-zero values for pixels which value for corresponding
		/// pixel in this image is greater than value for corresponding pixel in given image.
		QVImage<uChar> operator>(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise 'less or equal than' compare operator
		///
		/// @param img image to compare with this image.
		/// @returns 1-channel uChar bit depth image containing non-zero values for pixels which value for corresponding
		/// pixel in this image is less or equal than value for corresponding pixel in given image.
		QVImage<uChar> operator<=(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise 'greater or equal than' compare operator
		///
		/// @param img image to compare with this image.
		/// @returns 1-channel uChar bit depth image containing non-zero values for pixels which value for corresponding
		/// pixel in this image is greater or equal than value for corresponding pixel in given image.
		QVImage<uChar> operator>=(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise constant add operator.
		///
		/// Uses as the second operand a constant value for every pixel computed.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator+(const Type constant) const;

		/// @brief Pixel wise constant product operator.
		///
		/// Uses as the second operand a constant value for every pixel computed.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator*(const Type constant) const;

		/// @brief Pixel wise constant subtraction operator.
		///
		/// Uses as the second operand a constant value for every pixel computed.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator-(const Type constant) const;

		/// @brief Pixel wise constant division operator.
		///
		/// Uses as the second operand a constant value for every pixel computed.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator/(const Type constant) const;

		/// @brief Pixel wise bit-wise left shift operator.
		///
		/// Uses as the value to shift a constant value for every pixel computed.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator<<(const Type constant) const;

		/// @brief Pixel wise bit-wise right shift operator.
		///
		/// Uses as the value to shift a constant value for every pixel computed.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator>>(const Type constant) const;

		/// @brief Pixel wise bit-wise NOT operation.
		///
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator!() const;

		/// @brief Pixel wise bit-wise boolean and operator.
		///
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator&(const Type constant) const;

		/// @brief Pixel wise bit-wise boolean or operator.
		///
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator|(const Type constant) const;

		/// @brief Pixel wise bit-wise boolean xor operator.
		///
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param constant second operand value.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator^(const Type constant) const;

		/// @brief Pixel wise add operator.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param img operand image.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator+(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise mult operator.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param img operand image.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator*(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise subtraction operator.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param img operand image.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator-(const QVImage<Type, Planes> &img) const;

		/// @brief Pixel wise division operator.
		///
		/// Saturation is used for integer bit depth data (any except sFloat bit depth).
		/// Regions of interests, and anchors are applied in this operator.
		///
		/// @param img operand image.
		/// @returns resulting image for the operation.
		QVImage<Type, Planes> operator/(const QVImage<Type, Planes> &img) const;

	protected:
		QSharedDataPointer< QVImageBuffer<Type,Planes> > imageBuffer;
	};

template <typename Type, int C> bool QVImage<Type, C>::operator==(const QVImage<Type, C> &img) const
	{
	Q_ASSERT_X(img.getPlanes() == this->getPlanes(), "QVImage::operator==", "different number of planes");
	if (this->getCols() != img.getCols()) return false;
	if (this->getRows() != img.getRows()) return false;
	if (this->getPlanes() != img.getPlanes()) return false;
	if (this->getROI() != img.getROI()) return false;
	QVIMAGE_INIT_READ(Type,img);
	QVIMAGE_PTR_INIT_READ(Type,this);

	uInt x0 = (uInt) img.getROI().x(), y0 = (uInt) img.getROI().y(); 
	uInt x1 = (uInt) img.getROI().width(), y1 = (uInt) img.getROI().height(); 

	uInt lineSize = x1 * img.getPlanes();
	for(uInt row = y0; row < y1; row++)
		if (memcmp(&QVIMAGE_PIXEL(img, x0, row, 0),&QVIMAGE_PIXEL(this, x0, row, 0), lineSize) != 0)
			return false;	
	return true;
	};


typedef QVImage<uChar,1> QVImageUCharC1;
typedef QVImage<uChar,3> QVImageUCharC3;
typedef QVImage<sFloat,1> QVImageSFloatC1;
typedef QVImage<sFloat,3> QVImageSFloatC3;
typedef QVImage<sShort,1> QVImageSShortC1;
typedef QVImage<sShort,3> QVImageSShortC3;

//Q_DECLARE_METATYPE(QVGenericImage);
Q_DECLARE_METATYPE(QVImageUCharC1);
Q_DECLARE_METATYPE(QVImageUCharC3);
Q_DECLARE_METATYPE(QVImageSFloatC1);
Q_DECLARE_METATYPE(QVImageSFloatC3);
Q_DECLARE_METATYPE(QVImageSShortC1);
Q_DECLARE_METATYPE(QVImageSShortC3);

#endif // QVIMAGE_H
